#!/usr/bin/env python
##
## 4x4antiper.py
## Modification of the previous bonds_evo.py
## where the system is 4x4
## and bonds are laid in an antisymmetric,
## rather than symmetric, fashion.
##
## Spins are laid in a matrix with X rows and Y columns:
## if we count them row-by-row,
## then spin s[i][j] where row i and column j meet
## is the (i*Y+j)-th bit.
## North bond of spin s[i][j] is numbered 2*(i*X+j)
## West  bond of spin s[i][j] is numbered 2*(i*X+j)+1
## Spin s[i][j] may flip at even times if and only if i+j is even
## Spin s[i][j] may flip at odd  times if and only if i+j is odd
##
## Each spin controls two bonds: north and west
## As we said, spin s[i][j] is the (i*Y+j)-th bit:
## thus, it controls bonds 2*(i*Y+j) and 2*(i*Y+j)+1
## Antisymmetry works like this:
## bonds follow a scheme p q
##                       q p where q is the negation of p.
##
## 2011-01-24, 2011-01-31 Silvio Capobianco and Tommaso Toffoli

import time

X     = 4       ## number of rows    --- should be even
Y     = 4       ## number of columns --- should be even
SPINS = X*Y     ## number of spins
BONDS = SPINS*2 ## each site controls two bonds

def bit(i,n):
  '''Returns bit i of number n.
  Useful when one does not want to write "(n>>i)&1".'''
  return (n>>i)&1

## Binary printing function
def binstr(n,len=8):
  '''Writes the len least significant bits of n'''
  return ''.join(map(lambda t: str((n>>t)&1),range(len-1,-1,-1)))

def next_state(cfg,ph,lo):
  '''Compute next state of config cfg in phase ph,
  given layout of bonds lo.
  cfg is an integer between 0 and 2**SPINS-1
  ph  is a  bit
  lo  is an integer between 0 and 2**BONDS-1'''
  res = cfg
  for i in range(X):
    for j in range(Y):
      if (i+j-ph)%2 == 0:               ## if i+j has same parity as ph
        ce = i*Y+j                      ## index of center spin
        no = ce-Y if i>0   else X*Y-Y+j ## index of north neighbor
        so = ce+Y if i<X-1 else j       ## index of south neighbor
        ea = ce+1 if j<Y-1 else i*Y     ## index of east  neighbor
        we = ce-1 if j>0   else ce+Y-1  ## index of west  neighbor
        nb = 2*ce                       ## index of north bond
        sb = 2*so                       ## index of south bond
        eb = 2*ea+1                     ## index of east  bond
        wb = 2*ce+1                     ## index of west  bond
        sum = (bit(no,cfg) ^ bit(nb,lo)) \
            + (bit(so,cfg) ^ bit(sb,lo)) \
            + (bit(ea,cfg) ^ bit(eb,lo)) \
            + (bit(we,cfg) ^ bit(wb,lo))
        if sum==2: res ^= (1L << ce)    ## we flip bit ce in res
  return res

def neb(cfg,lo):
  '''Computes the number of excited bonds in configuration cfg'''
  res = 0
  for b in range(BONDS):
    s = b/2 ## site controlling b
    if b%2: ## if b is odd  (horizontal) check against west
      o = s-1 if (s%Y) else s+Y-1
    else:   ## if b is even (vertical)   check against north
      o = s-Y if (s>=Y) else s+X*Y-Y
    res += (bit(s,cfg) ^ bit(o,cfg) ^ bit(b,lo))
  return res

def bonds_4x4antisym(base):
  '''Construct 4x4 antisymmetric bonds
  by replicating base 2x2 structure'''
  lo    =  base       & 0xf ## bonds  0,    1,    2,    3
  hi    = (base >> 4) & 0xf ## bonds 2Y, 2Y+1, 2Y+2, 2Y+3
  loneg =  lo         ^ 0xf ## negation of lo
  hineg =  hi         ^ 0xf ## negation of hi
  res  = lo                 ## bonds 00, 01, 02, 03
  res += loneg * (1L <<  4) ## bonds 04, 05, 06, 07
  res += hi    * (1L <<  8) ## bonds 08, 09, 0a, 0b
  res += hineg * (1L << 12) ## bonds 0c, 0d, 0e, 0f
  res += loneg * (1L << 16) ## bonds 10, 11, 12, 13
  res += lo    * (1L << 20) ## bonds 14, 15, 16, 17
  res += hineg * (1L << 24) ## bonds 18, 19, 1a, 1b
  res += hi    * (1L << 28) ## bonds 1c, 1d, 1e, 1f
  return res

def construct_orbit(state,phase,layout):
  '''Construct orbit starting from given state
  in given phase and layout.
  An orbit is given as a list of (state,phase) 2-tuples.'''
  res = []
  cs = state
  cp = phase
  while True:
    res.append((cs,cp))
    cs,cp = next_state(cs,cp,layout),1-cp
    if (cs,cp) in res: break
  return res

def construct_dynamics(layout):
  '''Returns the list of orbits for the dynamics
  with the given layout of bonds.
  A dynamics is given as a list of orbits.'''
  res = []
  curr = [(state,phase) for phase in [0,1] for state in range(16)]
  while curr != []:
    cs,cp = curr[0] ## starting point of the dynamics
    orbit = construct_orbit(cs,cp,layout)
    for pair in orbit: curr.remove(pair)
    res.append(orbit)
  return res

def construct_classes(dyns):
  '''Construct the equivalence classes defined by stating
  that two layouts are equivalent iff they generate the same dynamics.
  Input is given as a dictionary of the form "layout:dynamics".
  Result is formatted as a list of lists,
  each made of a dynamics and the list of layouts that generate it.'''
  res = []
  for i in dyns.keys():
    toadd = 1
    for quid in res:
      if dyns[i]==quid[0]:
        quid[1].append(i)
        toadd = 0
    if toadd:
      res.append([dyns[i],[i]])
  return res

def display_config(cfg):
  '''Displays a single configuration as a matrix with X rows and Y columns'''
  for i in range(X):
    print '   '.join(map(str(cfg >> (i*Y+j)),range(Y-1,-1,-1)))


def display_step(cfg,ph,lo):
  '''Given an initial configuration, a phase, and a bonds layout,
  prints a table with layout, four semi-configs in a row,
  and total energy'''
  DISPLAY_FMT = '%s '*(2*X+1) + ' %3d'
  nxt = next_state(next_state(cfg,ph,lo),1-ph,lo)
  print DISPLAY_FMT % ((hexstr(lo,BONDS),) \
                         + tuple([binstr(cfg/(2**(i*Y)),Y) for i in range(X)]) \
                         + tuple([binstr(nxt/(2**(i*Y)),Y) for i in range(X)]) \
                         + (neb(cfg,lo),))

def display_classes():
  dyns = {}
  for i in range(256): dyns[i] = construct_dynamics(i)
  equiv = construct_classes(dyns)
  for c in equiv: print c

def test_display():
  for base in range(256):
    for config in range(16):
      display_step(config,0,bonds_2x2(base))
    print

def cool(layouts):
  '''Does the cool thing of the experiment'''
  Edic = {}				# Dict of Bonds & States vs Energies
  for i in layouts:
    Eple = []				# Energy tuple
    Sple = []
    for c in range(2**SPINS):           # every possible state of spins
      E,nxt = neb(c,i),next_state(next_state(c,0,i),1,i)
      Eple.append(E)
      Sple.append(nxt)
    Eple = tuple(Eple)                  # Convert lists to tuples so that
    Sple = tuple(Sple)                  # they can act as dictionary keys
    if Eple not in Edic: Edic[Eple] = []# Append bonds to Edic
    Edic[Eple].append((i,Sple))
  for E in Edic:			# for each possible Energy array
    print '---',
    for i in E: print i,		# print this energy array
    print
    for i,Sple in Edic[E]:	# extract Bonds structure, State arra
      print '%2x:' % i,		# print bond
      for j in Sple:		# print state
        print '%x' % j,
      print
    print

def construct_dyns_by_neb(layouts):
  '''Constructs the dynamics associated to a class of layouts
  as a dictionary whose keys are 2^SPINS-tuples of energies
  associated to states'''
  Edic = {}				# Dict of Bonds & States vs Energies
  for i in layouts:
    Eple = []				# Energy tuple
    Sple = []
    for c in range(2**SPINS):           # every possible state of spins
      E,nxt = neb(c,i),next_state(next_state(c,0,i),1,i)
      Eple.append(E)
      Sple.append(nxt)
    Eple = tuple(Eple)                  # Convert lists to tuples so that
    Sple = tuple(Sple)                  # they can act as dictionary keys
    if Eple not in Edic: Edic[Eple] = []# Append bonds to Edic
    Edic[Eple].append((i,Sple))
  return Edic

def display_dyns_by_neb(Edic):
  '''Displays the dynamics indicized by energy tuples in Edic'''
  for E in Edic:			# for each possible Energy array
    print '---',
    for i in E: print i,		# print this energy array
    print
    for i,Sple in Edic[E]:	# extract Bonds structure, State arra
      print '%2x:' % i,		# print bond
      for j in Sple:		# print state
        print '%x' % j,
      print
    print

def check_conj(Edic):
  '''Tests the Toffoli-Capobianco conjecture against dynamics
  in dictionary Edic generated by construct_dyns_by_neb()'''
  for e in Edic.values():
    for i in e[1:]:
      if i[1:] != e[0][1:]:
        print 'CONJECTURE DISPROVED!'
        print
        print e[0]
        print
        print i
        print
        return
  print 'Nothing odd, conjecture seems plausible for now...'
      

def all_equal(items):
  for i in items[1:]:
    if i != items[0]:
      return False
  return True

if __name__=='__main__':
  print 'Experiment started', time.ctime()
  print 'Constructing dynamics...'
##  Edic = construct_dyns_by_neb(range(2**BONDS))
  layouts = [bonds_4x4antisym(base) for base in range(256)]
  Edic = construct_dyns_by_neb(layouts)
  print 'Checking... '
  check_conj(Edic)
  print 'Experiment   ended', time.ctime()
##  test_display()
##  cool(range(256))
##  display_dyns_by_neb(construct_dyns_by_neb(range(2**BONDS)))
##  cool([bonds_2x2(base) for base in range(256)])

'''
enscript --margins=24:24:24:24 -Mletter -h -4 -fCourier\@5.6/7 -ops 2x2emap
'''



'''
Output should be done as such:
- list of all the energies
- list of all possible distinct spin configurations
- all the values of bonds corresponding to that spin config and energy list
'''
